/*:
 * @plugindesc タイルを書き換えるプラグイン
 * @target MZ
 * @base PluginCommonBase
 * @orderAfter PluginCommonBase
 * 
 * @command setAutoTile
 * @text タイルを変更
 * @desc 指定座標のタイルを書き換えます。
 *
 * @arg x
 * @type string
 * @text X座標
 * @desc タイルを変更するX座標
 *
 * @arg y
 * @type string
 * @text Y座標
 * @desc タイルを変更するY座標
 *
 * @arg z
 * @type string
 * @text Z座標
 * @desc タイルを変更するZ座標（0: 下層, 1: 中下層, 2: 中上層, 3: 上層）
 *
 * @arg tileId
 * @type string
 * @text タイルID
 * @desc 設定するタイルID
 * 
 * @arg isAutoTile
 * @type boolean
 * @text オートタイル
 * @desc オートタイルで周囲になじませる
 * タイルセットA1～A4までのものが対応
 *
 * @command updateAutoTile
 * @text オートタイルを更新
 * @desc 指定座標のタイルをオートタイルで馴染ませます。
 *
 * @arg x
 * @type string
 * @text X座標
 * @desc オートタイルを更新するX座標
 *
 * @arg y
 * @type string
 * @text Y座標
 * @desc オートタイルを更新するY座標
 *
 * @arg z
 * @type string
 * @text Z座標
 * @desc オートタイルを更新するZ座標（0: 下層, 1: 中下層, 2: 中上層, 3: 上層）
 * 
 * @command setNeighborsAutoTile
 * @text 周囲のオートタイルを変更
 * @desc 指定座標の周囲のタイルをオートタイルで馴染ませます。
 *
 * @arg Nx
 * @type string
 * @text X座標
 * @desc 中心となるタイルのX座標
 *
 * @arg Ny
 * @type string
 * @text Y座標
 * @desc 中心となるタイルのY座標
 *
 * @arg Nz
 * @type string
 * @text Z座標
 * @desc タイルを変更するZ座標（0: 下層, 1: 中下層, 2: 中上層, 3: 上層）
 * 
 * @command clearTileChanges
 * @text 指定マップのタイル変更をリセット
 * @desc 指定したマップIDのタイル変更情報をリセットします。
 * （1～: マップID, 0: 全てのマップ, -1: 現在のマップ）
 *
 * @arg mapId
 * @type string
 * @text マップID
 * @desc リセットするマップID
 * （1～: マップID, 0: 全てのマップ, -1: 現在のマップ）
 * 
 * @command checkTileId
 * @text タイルIDを取得
 * @desc 指定座標のタイルIDを取得し変数に格納します。
 *
 * @arg x
 * @type string
 * @text X座標
 * @desc 取得するタイルのX座標
 *
 * @arg y
 * @type string
 * @text Y座標
 * @desc 取得するタイルのY座標
 *
 * @arg z
 * @type string
 * @text Z座標
 * @desc 取得するタイルのZ座標（0: 下層, 1: 中下層, 2: 中上層, 3: 上層）
 * 
 * @arg variableId
 * @type variable
 * @text 変数ID
 * @desc 取得したタイルIDを格納する変数ID
 * 
 * @help
 * 指定座標のタイルをオートタイルを考慮し書き換えます。
 * 公式プラグインPluginCommonBase.jsをベースにしています。
 * 
 * 課題
 * ・再描画までにラグがある
 * ・履歴消去時のマップ再描画の改善
 * ・壁面タイルの詳細なオートタイルにまだ未着手
 * ・必要のない座標調査が多い
 */

(() => {
    const neighborsOffset = [
        { dx: 0, dy: 0 },                                        // 空データ
        { dx: -1, dy: 1 }, { dx: 0, dy: 1 }, { dx: 1, dy: 1 },   // 下
        { dx: -1, dy: 0 }, { dx: 0, dy: 0 }, { dx: 1, dy: 0 },   // 中
        { dx: -1, dy: -1 }, { dx: 0, dy: -1 }, { dx: 1, dy: -1 } // 上
    ];
    //オートタイルパターン表
    const autoTilePattern1 = [//上下左右に加え斜めも接続するタイプ　パターン全47通り
        [0, 1, 1, 1, 1, 1, 1, 1, 1, 1], [0, 1, 1, 1, 1, 1, 1, 0, 1, 1], [0, 1, 1, 1, 1, 1, 1, 1, 1, 0], [0, 1, 1, 1, 1, 1, 1, 0, 1, 0],//上下左右が繋がっているパターン
        [0, 1, 1, 0, 1, 1, 1, 1, 1, 1], [0, 1, 1, 0, 1, 1, 1, 0, 1, 1], [0, 1, 1, 0, 1, 1, 1, 1, 1, 0], [0, 1, 1, 0, 1, 1, 1, 0, 1, 0],//同上
        [0, 0, 1, 1, 1, 1, 1, 1, 1, 1], [0, 0, 1, 1, 1, 1, 1, 0, 1, 1], [0, 0, 1, 1, 1, 1, 1, 1, 1, 0], [0, 0, 1, 1, 1, 1, 1, 0, 1, 0],//同上
        [0, 0, 1, 0, 1, 1, 1, 1, 1, 1], [0, 0, 1, 0, 1, 1, 1, 0, 1, 1], [0, 0, 1, 0, 1, 1, 1, 1, 1, 0], [0, 0, 1, 0, 1, 1, 1, 0, 1, 0],//同上
        [0, 0, 1, 1, 0, 1, 1, 0, 1, 1], [0, 0, 1, 1, 0, 1, 1, 0, 1, 0], [0, 0, 1, 0, 0, 1, 1, 0, 1, 1], [0, 0, 1, 0, 0, 1, 1, 0, 1, 0],//左が繋がっていないパターン
        [0, 1, 1, 1, 1, 1, 1, 0, 0, 0], [0, 1, 1, 0, 1, 1, 1, 0, 0, 0], [0, 0, 1, 1, 1, 1, 1, 0, 0, 0], [0, 0, 1, 0, 1, 1, 1, 0, 0, 0],//上が繋がっていないパターン
        [0, 1, 1, 0, 1, 1, 0, 1, 1, 0], [0, 0, 1, 0, 1, 1, 0, 1, 1, 0], [0, 1, 1, 0, 1, 1, 0, 0, 1, 0], [0, 0, 1, 0, 1, 1, 0, 0, 1, 0],//右が繋がっていないパターン
        [0, 0, 0, 0, 1, 1, 1, 1, 1, 1], [0, 0, 0, 0, 1, 1, 1, 0, 1, 1], [0, 0, 0, 0, 1, 1, 1, 1, 1, 0], [0, 0, 0, 0, 1, 1, 1, 0, 1, 0],//下が繋がっていないパターン
        [0, 0, 1, 0, 0, 1, 0, 0, 1, 0], [0, 0, 0, 0, 1, 1, 1, 0, 0, 0],//左・右、上・下が繋がってないパターン
        [0, 0, 1, 1, 0, 1, 1, 0, 0, 0], [0, 0, 1, 0, 0, 1, 1, 0, 0, 0],//左・上が繋がってないパターン　┌
        [0, 1, 1, 0, 1, 1, 0, 0, 0, 0], [0, 0, 1, 0, 1, 1, 0, 0, 0, 0],//右・上が繋がってないパターン　┐
        [0, 0, 0, 0, 1, 1, 0, 1, 1, 0], [0, 0, 0, 0, 1, 1, 0, 0, 1, 0],//右・下が繋がってないパターン　┘
        [0, 0, 0, 0, 0, 1, 1, 0, 1, 1], [0, 0, 0, 0, 0, 1, 1, 0, 1, 0],//左・下が繋がってないパターン　└
        [0, 0, 1, 0, 0, 1, 0, 0, 0, 0], [0, 0, 0, 0, 0, 1, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0, 1, 0], [0, 0, 0, 0, 1, 1, 0, 0, 0, 0],//下、右、上、左以外繋がってないパターン
        [0, 0, 0, 0, 0, 1, 0, 0, 0, 0]//どことも繋がってないパターン
    ];
    const autoTilePattern2 = [//上下左右のみ接続するタイプ　パターン全16通り　最下位ビットから左・上・右・下の順で2進数方式
        [0, 0, 1, 0, 1, 1, 1, 0, 1, 0], [0, 0, 1, 0, 0, 1, 1, 0, 1, 0], [0, 0, 1, 0, 1, 1, 1, 0, 0, 0], [0, 0, 1, 0, 0, 1, 1, 0, 0, 0],
        [0, 0, 1, 0, 1, 1, 0, 0, 1, 0], [0, 0, 1, 0, 0, 1, 0, 0, 1, 0], [0, 0, 1, 0, 1, 1, 0, 0, 0, 0], [0, 0, 1, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 1, 1, 0, 1, 0], [0, 0, 0, 0, 0, 1, 1, 0, 1, 0], [0, 0, 0, 0, 1, 1, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 1, 0, 0, 0],
        [0, 0, 0, 0, 1, 1, 0, 0, 1, 0], [0, 0, 0, 0, 0, 1, 0, 0, 1, 0], [0, 0, 0, 0, 1, 1, 0, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],


    ];
    const autoTilePattern3 = [//左右のみ接続するタイプ　パターン全4通り
        [0, 0, 0, 0, 1, 1, 1, 0, 0, 0], [0, 0, 0, 0, 0, 1, 1, 0, 0, 0], [0, 0, 0, 0, 1, 1, 0, 0, 0, 0], [0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
    ];
    //const param = PluginManagerEx.createParameter(document.currentScript);
    PluginManagerEx.registerCommand(document.currentScript, "setAutoTile", args => {
        const x = Number(args.x);
        const y = Number(args.y);
        const z = Number(args.z);
        const isAutoTile = Boolean(args.isAutoTile);
        const tileId = Number(args.tileId);
        if (isAutoTile) {
            $gameMap.setAutoTile(x, y, z, tileId);//オートタイルON
        } else {
            $gameMap.setTile(x, y, z, tileId);//オートタイルOFF
        };
        $gameMap.requestRefresh();
    });
    PluginManagerEx.registerCommand(document.currentScript, "updateAutoTile", args => {
        const x = Number(args.x);
        const y = Number(args.y);
        const z = Number(args.z);
        $gameMap.updateAutoTile(x, y, z);
        $gameMap.requestRefresh();
    });
    PluginManagerEx.registerCommand(document.currentScript, "setNeighborsAutoTile", args => {
        const x = Number(args.Nx);
        const y = Number(args.Ny);
        const z = Number(args.Nz);
        neighborsOffset.forEach((offset, index) => {
            if (index === 0 || index === 5) { return; };// 空データと指定座標（中央）をスキップ
            const nx = x + offset.dx;
            const ny = y + offset.dy;
            if ($gameMap.isValid(nx, ny)) {
                $gameMap.updateAutoTile(nx, ny, z);
            };
        });
        $gameMap.requestRefresh();
    });
    PluginManagerEx.registerCommand(document.currentScript, "clearTileChanges", args => {
        const mapId = Number(args.mapId);
        if (mapId < 0) {
            $gameMap.clearTileChanges($gameMap.mapId());// 指定されたマップが-1の場合現在のマップIDに
        } else {
            $gameMap.clearTileChanges(mapId);
        };
        $gameMap.requestRefresh();
    });
    PluginManagerEx.registerCommand(document.currentScript, "checkTileId", args => {
        const x = Number(args.x);
        const y = Number(args.y);
        const z = Number(args.z);
        const variableId = Number(args.variableId);
        $gameVariables.setValue(variableId, $gameMap.tileId(x, y, z))

    });

    Game_Map.prototype.setAutoTile = function (x, y, z, tileId) {
        if (!this.isValid(x, y)) return; // 無効な座標は無視
        const tileId1 = this.tileId(x, y, z);
        const tileKind = Tilemap.getAutotileKind(tileId);
        const tileCategory = Classify(tileId);

        // 周囲8方向を調査
        const neighborTiles = [];
        neighborsOffset.forEach((offset, index) => {
            if (index === 0) {             // 空データをスキップ
                neighborTiles.push(0);
                return;
            };
            if (index === 5) {             // 指定座標（中央）をスキップ
                neighborTiles.push(1);
                return;
            };
            const nx = x + offset.dx;
            const ny = y + offset.dy;
            if (this.isValid(nx, ny)) {
                const a = $gameMap.tileId(nx, ny, z);
                var b;
                if (tileCategory === 1) {//水面の場合、滝の隣接も判定
                    b = Number(Tilemap.isWaterfallTile(a) || Tilemap.isSameKindTile(a, tileId))
                } else if (tileCategory === 4) {//滝の場合、水面・滝の隣接も判定
                    b = Number(Tilemap.isWaterTile(a));
                } else {
                    b = Number(Tilemap.isSameKindTile(a, tileId));
                };
                neighborTiles.push(b);
            } else {
                neighborTiles.push(0);
            };
        });
        //調査結果を整形しパターン表と照合
        const shape = reshape(neighborTiles, tileCategory);

        //タイルIDを再計算して変更の必要がある（現在のタイルとは違うタイル・パターンになる）場合更新
        const tileId3 = Tilemap.makeAutotileId(tileKind, shape);
        if (tileId1 !== tileId3) {
            this.setTile(x, y, z, tileId3);
        };

    };


    Game_Map.prototype.setTile = function (x, y, z, tileId) {
        if (!this.isValid(x, y)) return; // 無効な座標は無視

        const width = $dataMap.width;
        const height = $dataMap.height;
        const index = x + y * width + z * width * height;

        $dataMap.data[index] = tileId;  // タイルデータを書き換え

        // 履歴に記録
        const mapId = this.mapId();
        $gameSystem._tileChanges = $gameSystem._tileChanges || {};
        $gameSystem._tileChanges[mapId] = $gameSystem._tileChanges[mapId] || [];
        // 重複チェック（同じ座標の変更が既に記録されていれば更新）
        const changes = $gameSystem._tileChanges[mapId];
        const existingChange = changes.find(change => change.x === x && change.y === y && change.z === z);
        if (existingChange) {
            existingChange.tileId = tileId;
        } else {
            changes.push({ x, y, z, tileId });
        }

        this.requestRefresh();          // マップをリフレッシュ
    };

    //指定座標がオートタイルか判別し、オートタイルなら更新
    Game_Map.prototype.updateAutoTile = function (x, y, z) {
        if (!this.isValid(x, y)) return;
        const tileId = this.tileId(x, y, z);
        const isAutoTile = Tilemap.isAutotile(tileId);
        if (isAutoTile) {
            this.setAutoTile(x, y, z, tileId);
        };
    };

    // マップ読み込み時に履歴を元に変更を適用
    Game_Map.prototype.applyTileChanges = function () {
        const mapId = this.mapId();
        const changes = $gameSystem._tileChanges?.[mapId];
        if (changes) {
            changes.forEach(change => {
                const width = $dataMap.width;
                const height = $dataMap.height;
                const index = change.x + change.y * width + change.z * width * height;
                $dataMap.data[index] = change.tileId;
            });
        }
    };
    const _Scene_Map_onMapLoaded = Scene_Map.prototype.onMapLoaded;
    Scene_Map.prototype.onMapLoaded = function () {
        _Scene_Map_onMapLoaded.call(this);
        $gameMap.applyTileChanges();
    };

    //履歴の消去
    Game_Map.prototype.clearTileChanges = function (mapId) {
        if (mapId === 0) {//指定されたマップが0の場合全消去
            delete $gameSystem._tileChanges;
            SceneManager.goto(Scene_Map);
        } else if ($gameSystem._tileChanges?.[mapId]) {
            delete $gameSystem._tileChanges[mapId];
            // その場でリフレッシュするのに適当な処理が思いつかない
            if (mapId === $gameMap.mapId()) {
                SceneManager.goto(Scene_Map);
            };
        }
    };

    function Classify(tileId) {
        if (Tilemap.isWaterTile(tileId)) {
            if (Tilemap.isWaterfallTile(tileId)) {
                return 4;//autoTilePattern3_滝
            } else {
                return 1;//autoTilePattern1_水面
            };
        } else if (Tilemap.isWallSideTile(tileId)) {
            return 3;    //autoTilePattern2_壁側面
        } else if (Tilemap.isRoofTile(tileId)) {
            return 2;    //autoTilePattern2_屋根
        } else {
            return 0;    //autoTilePattern1_地面等
        };
    };

    function reshape(neighborTiles, tileCategory) {
        if (tileCategory < 0) { console.warn("整形エラー1"); return };
        if (tileCategory <= 1) { //パターン1の場合、上下左右の繋がっている面に合わせて斜めの情報を塗りつぶし
            if (!neighborTiles[2]) {
                neighborTiles[1] = 0;
                neighborTiles[3] = 0;
            };
            if (!neighborTiles[4]) {
                neighborTiles[1] = 0;
                neighborTiles[7] = 0;
            };
            if (!neighborTiles[6]) {
                neighborTiles[3] = 0;
                neighborTiles[9] = 0;
            };
            if (!neighborTiles[8]) {
                neighborTiles[7] = 0;
                neighborTiles[9] = 0;
            }
            //パターン表と照合
            const shape = autoTilePattern1.findIndex(element => element.every((value, index) => value === neighborTiles[index]));
            if (shape < 0) { console.warn("パターン照合エラー1"); return; };//エラーチェック
            return shape;
        } else if (tileCategory <= 3) {//パターン2の場合、斜めの情報を塗りつぶし
            neighborTiles[1] = 0;
            neighborTiles[3] = 0;
            neighborTiles[7] = 0;
            neighborTiles[9] = 0;
            //パターン表と照合
            const shape = autoTilePattern2.findIndex(element => element.every((value, index) => value === neighborTiles[index]));
            if (shape < 0) { console.warn("パターン照合エラー2"); return; };//エラーチェック
            return shape;
        } else if (tileCategory === 4) {//パターン3の場合、横以外の情報を塗りつぶし
            neighborTiles[1] = 0;
            neighborTiles[2] = 0;
            neighborTiles[3] = 0;
            neighborTiles[7] = 0;
            neighborTiles[8] = 0;
            neighborTiles[9] = 0;
            //パターン表と照合
            const shape = autoTilePattern3.findIndex(element => element.every((value, index) => value === neighborTiles[index]));
            if (shape < 0) { console.warn("パターン照合エラー3"); return; };//エラーチェック
            return shape;
        } else {
            console.warn("整形エラー2");
        }
    };
})();